@@ -192,6 +192,22 @@ module Test
         assert obj.nil?, msg
       end
 
+      ##
+      # Fails unless +obj+ is true
+
+      def assert_true obj, msg = nil
+        msg = message(msg) { "Expected #{mu_pp(obj)} to be true" }
+        assert obj == true, msg
+      end
+
+      ##
+      # Fails unless +obj+ is false
+
+      def assert_false obj, msg = nil
+        msg = message(msg) { "Expected #{mu_pp(obj)} to be false" }
+        assert obj == false, msg
+      end
+
       ##
       # For testing with binary operators.
       #
@@ -547,10 +563,6 @@ module Test
         assert yield, *msgs
       end
 
-      def assert_raises(*exp, &b)
-        raise NoMethodError, "use assert_raise", caller
-      end
-
       # :call-seq:
       #   assert_nothing_thrown( failure_message = nil, &block )
       #
@@ -756,7 +768,14 @@ EOT
           e = assert_raise(SyntaxError, mesg) do
             syntax_check(src, fname, line)
           end
-          assert_match(error, e.message, mesg)
+
+          # Prism adds ANSI escape sequences to syntax error messages to
+          # colorize and format them. We strip them out here to make them easier
+          # to match against in tests.
+          message = e.message
+          message.gsub!(/\e\[.*?m/, "")
+
+          assert_match(error, message, mesg)
           e
         end
       end
